MODULE PipeFork;

        (********************************************************)
        (*                                                      *)
        (*        Reads the piped output of the Weasel          *)
        (*   transaction log, sends the POP and SMTP client     *)
        (*  log entries out to two other pipes, and displays    *)
        (*           what's left on the screen.                 *)
        (*                                                      *)
        (*    Started:        20 April 2004                     *)
        (*    Last edited:    28 April 2004                     *)
        (*    Status:         OK                                *)
        (*                                                      *)
        (********************************************************)

IMPORT OS2, TextIO, STextIO, Strings;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM IOChan IMPORT
    (* type *)  ChanId;

FROM SYSTEM IMPORT
    (* proc *)  ADR;

FROM LowLevel IMPORT
    (* proc *)  AddOffset, ALLOCATE64;

FROM Storage IMPORT
    (* proc *)  DEALLOCATE;

(************************************************************************)

CONST
    InPipeName = "\\PIPE\\WeaselTransLog";
    PipeSName = "\\PIPE\\WeaselSMTPLog";
    PipePName = "\\PIPE\\WeaselPOPLog";
    Nul = CHR(0);
    PipeSize = 4096;
    LineBufferSize = 128;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Writes N in decimal to standard output. *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        STextIO.WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)

PROCEDURE CopyOver (inpipe, PipeS, PipeP: OS2.HPIPE);

    (* Reads all lines from inpipe, sends them to one of two output     *)
    (* pipes or to the screen, depending on category.                   *)

    CONST
        InBufferSize = 4096;

    VAR InBuffer: ARRAY [0..InBufferSize-1] OF CHAR;
        LineBuffer: ARRAY [0..LineBufferSize-1] OF CHAR;
        CharsInBuffer: CARDINAL;

    (********************************************************************)

    PROCEDURE GetOneLine(): CARDINAL;

        (* Reads one line of data from inpipe to LineBuffer.            *)
        (* Returns the number of characters in the line, including the  *)
        (* CRLF at the end of the line.                                 *)

        TYPE TwoChar = ARRAY [0..1] OF CHAR;

        CONST LF = CHR(10);  CR = CHR(13);
              CRLF = TwoChar {CR, LF};

        VAR count: CARDINAL;
            found: BOOLEAN;

        BEGIN
            REPEAT
                found := FALSE;
                count := 0;

                IF CharsInBuffer > 1 THEN

                    (* Find the first end-of-line. *)

                    Strings.FindNext (CRLF, InBuffer, 0, found, count);
                    IF found THEN
                        INC (count, 2);
                    ELSE
                        count := 0;
                    END (*IF*);
                END (*IF*);

                (* Read more from pipe if we're running short of data. *)

                IF (NOT found) AND (CharsInBuffer < LineBufferSize) THEN
                    IF OS2.DosRead (inpipe,
                                     AddOffset(ADR(InBuffer),CharsInBuffer),
                                      InBufferSize - CharsInBuffer,
                                       count) = 0 THEN
                        INC (CharsInBuffer, count);
                        IF CharsInBuffer < InBufferSize THEN
                            InBuffer[CharsInBuffer] := Nul;
                        END (*IF*);
                    END (*IF*);
                END (*IF*);

            UNTIL found OR (count = 0);

            IF NOT found THEN
                count := CharsInBuffer;
                IF count < 2 THEN
                    count := 0;
                ELSE
                    IF count > LineBufferSize THEN
                        count := LineBufferSize;
                    END (*IF*);
                    DEC (count, 2);
                END (*IF*);
            END (*IF*);

            IF count > 0 THEN

                (* Extract the line. *)

                Strings.Extract (InBuffer, 0, count, LineBuffer);
                Strings.Delete (InBuffer, 0, count);
                DEC (CharsInBuffer, count);

                (* If we have a truncated line, add a CRLF. *)

                IF NOT found THEN
                    LineBuffer[count] := CR;
                    LineBuffer[count+1] := LF;
                END (*IF*);

            END (*IF*);

            RETURN count;

        END GetOneLine;

    (********************************************************************)

    VAR amount: CARDINAL;
        code: ARRAY [0..1] OF CHAR;

    BEGIN
        (* Initial priming of InBuffer. *)

        IF OS2.DosRead (inpipe, ADR(InBuffer),
                                   InBufferSize, CharsInBuffer) <> 0 THEN
            CharsInBuffer := 0;
        END (*IF*);
        IF CharsInBuffer < InBufferSize THEN
            InBuffer[CharsInBuffer] := Nul;
        END (*IF*);
        REPEAT
            amount := GetOneLine();
            IF amount > 0 THEN

                (* Extract the two characters that let us check  *)
                (* the type of the log entry.                    *)

                Strings.Extract (LineBuffer, 20, 2, code);
                IF code[1] = ' ' THEN
                    IF code[0] = 'S' THEN
                        OS2.DosWrite (PipeS, ADR(LineBuffer), LENGTH(LineBuffer), amount);
                    ELSIF code[0] = 'P' THEN
                        OS2.DosWrite (PipeP, ADR(LineBuffer), LENGTH(LineBuffer), amount);
                    ELSE
                        STextIO.WriteString (LineBuffer);
                    END (*IF*);
                ELSE
                    STextIO.WriteString (LineBuffer);
                END (*IF*);
            END (*IF*);
        UNTIL amount = 0;
    END CopyOver;

(************************************************************************)

PROCEDURE DoTheJob;

    CONST
        BufferSize = 256;
        OpenFlags = OS2.OPEN_ACTION_FAIL_IF_NEW
                    + OS2.OPEN_ACTION_OPEN_IF_EXISTS;
        Mode = OS2.OPEN_FLAGS_FAIL_ON_ERROR + OS2.OPEN_FLAGS_SEQUENTIAL
                + OS2.OPEN_SHARE_DENYNONE
                + OS2.OPEN_FLAGS_NOINHERIT
                + OS2.OPEN_ACCESS_READONLY;

    VAR rc, Action: CARDINAL;
        hpipe, PipeS, PipeP: OS2.HPIPE;
        Waiting, success: BOOLEAN;

    BEGIN
        (* Create the two output pipes. *)

        rc := OS2.DosCreateNPipe (PipeSName, PipeS, OS2.NP_ACCESS_OUTBOUND,
                                                         1, PipeSize, 0, 0);
        success := rc = 0;
        IF success THEN
            rc := OS2.DosConnectNPipe (PipeS);
            success := (rc = 0) OR (rc = 233);
        END (*IF*);
        IF NOT success THEN
            STextIO.WriteString ("Failed to open SMTP output pipe, error code ");
            WriteCard (rc);
            STextIO.WriteLn;
        END (*IF*);

        rc := OS2.DosCreateNPipe (PipePName, PipeP, OS2.NP_ACCESS_OUTBOUND,
                                                         1, PipeSize, 0, 0);
        success := rc = 0;
        IF success THEN
            rc := OS2.DosConnectNPipe (PipeP);
            success := (rc = 0) OR (rc = 233);
        END (*IF*);
        IF NOT success THEN
            STextIO.WriteString ("Failed to open POP output pipe, error code ");
            WriteCard (rc);
            STextIO.WriteLn;
        END (*IF*);

        REPEAT
            rc := OS2.DosOpen (InPipeName, hpipe, Action, 0, 0, OpenFlags, Mode, NIL);
            Waiting := rc = 3;
            IF Waiting THEN
                STextIO.WriteString ("Waiting for input pipe to be activated");
                STextIO.WriteLn;
                OS2.DosSleep (200);
            END (*IF*);
        UNTIL NOT Waiting;
        IF rc = 0 THEN
            CopyOver (hpipe, PipeS, PipeP);
            OS2.DosClose (hpipe);
            STextIO.WriteString ("*** END OF DATA FROM PIPE ***");
            STextIO.WriteLn;
        ELSE
            STextIO.WriteString ("Failed to open input pipe, error code ");
            WriteCard (rc);
            STextIO.WriteLn;
        END (*IF*);
        STextIO.WriteString ("Finished.");
        STextIO.WriteLn;
    END DoTheJob;

(************************************************************************)
(*                           INITIALISATION                             *)
(************************************************************************)

PROCEDURE SetScreenSize (NumberOfRows, NumberOfCols: CARDINAL);

    (* Sets the number of rows and columsns in the VIO window. *)

    VAR vioModeInfoPtr: POINTER TO OS2.VIOMODEINFO;

    BEGIN
        ALLOCATE64 (vioModeInfoPtr, SIZE(OS2.VIOMODEINFO));
        vioModeInfoPtr^.cb := SIZE (OS2.VIOMODEINFO);
        OS2.VioGetMode (vioModeInfoPtr^, 0);
        vioModeInfoPtr^.row := NumberOfRows;
        vioModeInfoPtr^.col := NumberOfCols;
        OS2.VioSetMode (vioModeInfoPtr^, 0);
        DEALLOCATE (vioModeInfoPtr, SIZE(OS2.VIOMODEINFO));
    END SetScreenSize;

(************************************************************************)

PROCEDURE ProcessArguments;

    (* Reads program arguments "rows cols" (both optional),   *)
    (* sets the screen size.                                  *)

    VAR Buffer: ARRAY [0..LineBufferSize-1] OF CHAR;
        pos: CARDINAL;

    (********************************************************************)

    PROCEDURE SkipBlanks;

        (* Updates pos to skip over blanks in Buffer. *)

        BEGIN
            WHILE Buffer[pos] = ' ' DO
                INC (pos);
            END (*WHILE*);
        END SkipBlanks;

    (********************************************************************)

    PROCEDURE GetNumber (VAR (*INOUT*) N: CARDINAL);

        (* Reads an unsigned number starting at Buffer[pos], updates    *)
        (* pos.  N is left unchanged if there is no number there.       *)

        TYPE CharSet = SET OF CHAR;
        CONST Digits = CharSet{'0'..'9'};

        BEGIN
            IF Buffer[pos] IN Digits THEN
                N := 0;
                WHILE Buffer[pos] IN Digits DO
                    N := 10*N + (ORD(Buffer[pos]) - ORD('0'));
                    INC (pos);
                END (*WHILE*);
            END (*IF*);
        END GetNumber;

    (********************************************************************)

    VAR NumberOfRows, NumberOfCols: CARDINAL;
        args: ChanId;

    BEGIN                              (* Body of ProcessArguments *)
        args := ArgChan();
        IF IsArgPresent() THEN
            TextIO.ReadString (args, Buffer);
        ELSE
            Buffer := "";
        END (*IF*);
        pos := 0;

        NumberOfRows := 25;
        NumberOfCols := 80;
        SkipBlanks;
        GetNumber (NumberOfRows);
        SkipBlanks;
        GetNumber (NumberOfCols);

        (* Set the VIO screen size. *)

        SetScreenSize (NumberOfRows, NumberOfCols);

    END ProcessArguments;

(************************************************************************)

BEGIN
    ProcessArguments;
    DoTheJob;
END PipeFork.

